# -*- coding: utf-8 -*-

from enum import Enum
from typing import Dict, List, Tuple

import arrow
import numpy as np
from httpx import AsyncClient, URL
from loguru import logger

from app.core.config import settings
from app.services.service import Service
from app.utils.date import get_time_str, TIME_FMT
from app.utils.math import round_half_up


class InfoCode(str, Enum):
    temperature = 'Tdb'
    co2 = 'CO2'
    hcho = 'HCHO'
    pm2d5 = 'PM2d5'
    humidity = 'RH'
    supply_air_flow = 'SupplyAirFlow'
    supply_air_temperature = 'SupplyAirTemp'
    supply_air_temperature_set = 'SupplyAirTempSet'
    fan_speed = 'FanGear'
    fan_freq = 'FanFreq'
    fan_freq_set = 'FanFreqSet'
    supply_static_press = 'SupplyStaticPress'
    supply_static_press_set = 'SupplyStaticPressSet'
    running_status = 'RunStatus'
    cloud_status = 'InCloudStatus'
    equip_switch_set = 'EquipSwitchSet'


class DataPlatformService(Service):

    def __init__(
            self,
            client: AsyncClient,
            project_id: str,
            server_settings=settings
    ):
        super(DataPlatformService, self).__init__(client)
        self._project_id = project_id
        self._base_url = URL(server_settings.PLATFORM_HOST)
        self._now_time = get_time_str()
        self._secret = server_settings.PLATFORM_SECRET

    def _common_parameters(self) -> Dict:
        return {'projectId': self._project_id, 'secret': self._secret}

    async def get_realtime_data(self, code: InfoCode, object_id: str) -> float:
        url = self._base_url.join('data-platform-3/hisdata/query_by_obj')
        params = self._common_parameters()
        start_time = get_time_str(60 * 60, flag='ago')
        payload = {
            'criteria': {
                'id': object_id,
                'code': code.value,
                'receivetime': {
                    '$gte': start_time,
                    '$lte': self._now_time,
                }
            }
        }
        raw_info = await self._post(url, params, payload)

        try:
            latest_data = raw_info.get('Content')[-1].get('data')
            latest_time = raw_info.get('Content')[-1].get('receivetime')
            if arrow.get(latest_time, TIME_FMT).shift(minutes=15) < arrow.get(self._now_time, TIME_FMT):
                logger.info(f'delayed data - {object_id}: ({latest_time}, {latest_data})')
            value = round_half_up(latest_data, 2)
        except KeyError and IndexError:
            value = np.NAN
        except TypeError:
            value = -1.0

        return value

    async def get_duration(self, code: InfoCode, object_id: str, duration: int) -> List[Dict]:
        url = self._base_url.join('data-platform-3/hisdata/query_by_obj')
        params = self._common_parameters()
        start_time = get_time_str(duration, flag='ago')
        payload = {
            'criteria': {
                'id': object_id,
                'code': code.value,
                'receivetime': {
                    '$gte': start_time,
                    '$lte': self._now_time,
                }
            }
        }
        raw_info = await self._post(url, params, payload)

        try:
            content = raw_info.get('Content')
            latest_time = content[-1].get('receivetime')
            if arrow.get(latest_time, TIME_FMT).shift(minutes=15) < arrow.get(self._now_time, TIME_FMT):
                result = []
                logger.info(f'delayed data - {object_id}: ({latest_time})')
            else:
                result = [
                    {
                        'timestamp': item['receivetime'],
                        'value': item['data']
                    }
                    for item in content
                ]
        except KeyError:
            result = []

        return result

    async def get_past_data(self, code: InfoCode, object_id: str, interval: int) -> float:
        """
        Query past data from data platform.
        :param code: Info code
        :param object_id:
        :param interval: time interval(seconds) from now to past
        :return: a past value
        """
        url = self._base_url.join('data-platform-3/hisdata/query_by_obj')
        params = self._common_parameters()
        start_time = get_time_str(60 * 60 + interval, flag='ago')
        end_time = get_time_str(interval, flag='ago')
        payload = {
            'criteria': {
                'id': object_id,
                'code': code.value,
                'receivetime': {
                    '$gte': start_time,
                    '$lte': end_time,
                }
            }
        }
        raw_info = await self._post(url, params, payload)

        try:
            latest_data = raw_info.get('Content')[-1].get('data')
            latest_time = raw_info.get('Content')[-1].get('receivetime')
            if arrow.get(latest_time, TIME_FMT).shift(minutes=15) < arrow.get(end_time, TIME_FMT):
                logger.info(f'delayed data - {object_id}: ({latest_time}, {latest_data})')
            value = round_half_up(latest_data, 2)
        except KeyError and IndexError:
            value = np.NAN
        except TypeError:
            value = -1.0

        return value

    async def get_realtime_temperature(self, space_id: str) -> float:
        return await self.get_realtime_data(InfoCode.temperature, space_id)

    async def get_past_temperature(self, space_id: str, interval: int) -> float:
        return await self.get_past_data(InfoCode.temperature, space_id, interval)

    async def get_realtime_co2(self, space_id: str) -> float:
        return await self.get_realtime_data(InfoCode.co2, space_id)

    async def get_realtime_hcho(self, space_id: str) -> float:
        return await self.get_realtime_data(InfoCode.hcho, space_id)

    async def get_realtime_pm2d5(self, space_id: str) -> float:
        return await self.get_realtime_data(InfoCode.pm2d5, space_id)

    async def get_realtime_humidity(self, space_id: str) -> float:
        return await self.get_realtime_data(InfoCode.humidity, space_id)

    async def get_realtime_supply_air_flow(self, equipment_id: str) -> float:
        return await self.get_realtime_data(InfoCode.supply_air_flow, equipment_id)

    async def get_realtime_supply_air_temperature(self, equipment_id: str) -> float:
        return await self.get_realtime_data(InfoCode.supply_air_temperature, equipment_id)

    async def get_realtime_supply_air_temperature_set(self, equipment_id: str) -> float:
        return await self.get_realtime_data(InfoCode.supply_air_temperature_set, equipment_id)

    async def get_fan_speed(self, equipment_id: str) -> float:
        return await self.get_realtime_data(InfoCode.fan_speed, equipment_id)

    async def get_static_info(self, code: str, object_id: str):
        url = self._base_url.join('data-platform-3/object/batch_query')
        params = self._common_parameters()
        payload = {
            'customInfo': True,
            'criterias': [
                {
                    'id': object_id
                }
            ]
        }
        raw_info = await self._post(url, params, payload)

        try:
            info = raw_info['Content'][0]['infos'][code]
        except KeyError as e:
            logger.error(f'id: {object_id}, details: {e}')
            info = None

        return info

    async def get_air_flow_limit(self, equipment_id: str) -> Tuple[float, float]:
        lower = await self.get_static_info('MinAirFlow', equipment_id)
        upper = await self.get_static_info('MaxAirFlow', equipment_id)
        if not lower:
            lower = 150.0
        if not upper:
            upper = 2000.0

        return lower, upper

    async def get_schedule(self, equipment_id: str) -> Tuple[str, str]:
        on_time = await self.get_static_info('ctm-OnTime', equipment_id)
        off_time = await self.get_static_info('ctm-OffTime', equipment_id)
        if not on_time:
            on_time = '080000'
        if not off_time:
            off_time = '190000'

        return on_time, off_time

    async def get_realtime_fan_freq_set(self, equipment_id: str) -> float:
        return await self.get_realtime_data(InfoCode.fan_freq_set, equipment_id)

    async def get_realtime_supply_static_press(self, system_id: str) -> float:
        return await self.get_realtime_data(InfoCode.supply_static_press, system_id)

    async def get_realtime_supply_static_press_set(self, system_id: str) -> float:
        return await self.get_realtime_data(InfoCode.supply_static_press_set, system_id)

    async def get_realtime_running_status(self, equipment_id: str) -> float:
        return await self.get_realtime_data(InfoCode.running_status, equipment_id)

    async def get_cloud_status(self, equipment_id: str) -> float:
        return await self.get_realtime_data(InfoCode.cloud_status, equipment_id)

    async def set_code_value(self, object_id: str, code: InfoCode, value: float):
        url = self._base_url.join('data-platform-3/parameter/setting')
        params = self._common_parameters()
        payload = {
            'id': object_id,
            'code': code.value,
            'value': value
        }

        await self._post(url, params, payload)

    async def get_items_by_category(self, code) -> List:
        url = self._base_url.join('data-platform-3/object/subset_query')
        params = self._common_parameters()
        payload = {
            'criteria': {
                'type': [code]
            }
        }

        raw_info = await self._post(url, params, payload)
        items_id_list = [item['id'] for item in raw_info['Content']]

        return items_id_list
